#This software is a rerelease of CyberTOR_WY.R module of Microarray Expression
#Potential Suite with alterations in the input and ouput format in order to
#improve transition between modules of eMESS.

#CyberT algorithm from:
#CyberT for use with spotted microaray data based on the derivation in
#"Improved Statistical Inference from DNA Microarray Data Using Analysis of
#Variance and Bayesian Statistical Framework" A. D. Long et al, J. Bio. Chem.,
#276(22), pp 19937-19944, 2001.
#Westfall-Young Stepdown Permutation Algorithm from:
#Westfall-Young Stepdown Permutation Algorithm for use with microarray data
#based on the derivation in "Differentially Expressed Genes in Replicate cDNA 
#Microarray Experiments", S. Duroit et al, Technical Report #578, August 2000.
#http://stat-www.berkeley.edu/users/terry/zarray/Html/papersindex.html

#No part of this text or software may be used or reproduced in any matter
#whatsoever without written permission, except in the case of brief quotations
#embodied in critical articles of or for reviews.  For information address
#Brian W. Kirk (bwkbem@yahoo.com).
#Software Copyright (C) October 2002  Brian W. Kirk
#Form/Text Copyright (C) October 2002  Brian W. Kirk
#Released for use with R v1.5.1 (C)

#Copyright (C) October 2002  Brian W. Kirk
#Released for R v1.5.1 (C)

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

###############################################################################
#                                FUNCTIONS                                    #
###############################################################################

library(MASS);

#FUNCTION THAT CALCULATES FOR A VECTOR OF SIGNALS A ROBUST SUMMARY USING THE
#HUBER M-ESTIMATOR AND THE MAD ESTIMATOR FOR SCALE.  RESULTS ARE COMBINED INTO
#A DATA.FRAME, WITH INCOMPATIBLE SETS (IE WITH ONLY 1 ENTRY) AMENDED TO THE END
#OF THE DATA.FRAME.

  fhuber.summarize <- function(SignalVec=vec, Filter=filter,
                               NameIdentifier=name, HuberConf=k, HuberTol=tol)
  {
   #Prime Header of Output Table
   huber.report.template <- c("ProbeID", "mu", "s", "probes");
   
   #Determine the Number of Probes Present for each Probeset 
   probes <- NameIdentifier[Filter];
   probesetf <- factor(probes);
   probesets <- levels(probesetf);

   probe.number <- NA;
   for (i in 1:length(probesets)) {
      temp  <- length(probes[as.logical(match(
                            probes, probesets[i], nomatch=0))]);
      probe.number <- c(probe.number, temp)  
   }
   probe.number <- probe.number[2:(length(probe.number))];
   
   probe.number.table <- data.frame(cbind(probesets, probe.number))

   #Select Porbesets With More Than 1 Probe For Huber Analysis
   huber.ok <- probesets[which(as.numeric(
                                     as.vector(probe.number.table[,2])) > 1)];

   #Apply Estimators To Qualified Probesets
   if (length(huber.ok) > 0) {
     for (i in (1:length(huber.ok))) {

       #Test That More Than Half of the Probes Are Not Identical
       test.probe <- SignalVec[as.logical(
                        match(NameIdentifier, huber.ok[i], nomatch=0)*Filter)];
     
       test.probef <- factor(test.probe);
       test.probes <- levels(test.probef);
       level.lengths <- tapply(test.probe, test.probef, length);

       continue <- TRUE;
       for (j in 1:length(test.probes)) {
         if ((length(test.probe) - level.lengths[j]) < 0.5*length(test.probe)){
           continue <- FALSE
         }
       }

       #Process Accordingly Based on the Previous Test
       if (continue) {
         xtemp.matrix <-  as.matrix(huber(test.probe, k=HuberConf,
                                          tol=HuberTol));
         huber.report.template <- cbind(huber.report.template,
                        c(as.character(huber.ok[i]),
                          as.numeric(xtemp.matrix[1,1]),
                          as.numeric(xtemp.matrix[2,1]),
                          as.numeric(as.vector(
                              probe.number.table[as.logical(match(
                              probesets, huber.ok[i], nomatch=0)),2]))));
       } else {
         mu <- median(test.probe);
         scale <- sd(test.probe);

         huber.report.template <- cbind(huber.report.template,
                        c(as.character(huber.ok[i]),
                          mu,
                          scale,
                          as.numeric(as.vector(
                              probe.number.table[as.logical(match(
                              probesets, huber.ok[i], nomatch=0)),2]))));
   
       } 
     }
   }
   
   #Organize Porbesets That Did Not Qualify For Estimator Analysis
   huber.lost <- probesets[which(as.numeric(
                                    as.vector(probe.number.table[,2])) <= 1)];
   if (length(huber.lost) > 0) {
     for (i in (1:length(huber.lost))) {
       huber.report.template <- cbind(huber.report.template,
                            c(as.character(huber.lost[i]),
                              as.numeric(SignalVec
                                          [as.logical(match(NameIdentifier,
                                           huber.lost[i], nomatch=0)
                                           * Filter)]), NA, 1))
     }
   }

   #Combine Both Probeset Types Into A Single Report
   huber.report.template2 <- t(huber.report.template);
   huber.report.template3 <-
                        huber.report.template2[2:(length(probesets)+1),];
   huber.sort.filter <- order(as.numeric(huber.report.template3[,4]),
                               decreasing=TRUE );
   return(data.frame(cbind(
                  ProbeID=
                    as.character(huber.report.template3[huber.sort.filter, 1]),
                  mu=
                    as.numeric(huber.report.template3[huber.sort.filter, 2]),
                  s=
                    as.numeric(huber.report.template3[huber.sort.filter, 3]),
                  probes=
                    as.numeric(
                    as.vector(huber.report.template3[huber.sort.filter, 4])))))
  }


###############################################################################
#                           CONSTANT DECLARATION                              #
###############################################################################

#CONSTANTS AND VARIABLES OF TEMPORARY USE BEGIN WITH "x, y, or z" AND ALL
#USER DEFINED FUNCTIONS BEGIN WITH "f" or "g" IN ORDER TO FACILITATE REMOVABLE
#FROM MEMORY.  WHEN USING VARIABLES WITH THESE RESERVED INTIAL LETTERS,
#NEAREST CLEANUP SHOULD BE IDENTIFIED TO INSURE PROPER PERMANENCY

                 ############################################
                 #######  DEFINE PERMANENT CONSTANTS  #######
                 ############################################

#Define Output Dirctory with Prefix
OutputPrefix <- paste(OutputDirectory, OutputFile, sep="");

#DEFINE GRAPHICS WINDOW SIZE
WinWidth <- 8.5;           
WinHeight <- 11;


             #####################################################
             ##########  DEFINE TEMPORARY CONSTANTS  #############
             #####################################################


#Column position in LNProbesetDatawBGSubtract.RData
xIDPos <- 1;
MuFoldPos <- 2;
xMuAmpPos <- 3;
xProbeNumberPos <- 4;


###############################################################################
#                             VARIABLE DECLARATION                            #
###############################################################################

#CONSTANTS AND VARIABLES OF TEMPORARY USE BEGIN WITH "x", "y", or "z" AND ALL
#USER DEFINED FUNCTIONS BEGIN WITH "f" or "g" IN ORDER TO FACILITATE REMOVABLE
#FROM MEMORY.  WHEN USING VARIABLES WITH THESE RESERVED INTIAL LETTERS,
#NEAREST CLEANUP SHOULD BE IDENTIFIED TO INSURE PROPER PERMANCY



#Load LNProbesetData for each array into a list of data.frames
TotalArrayNumber <- length(ProbesetFiles);
ArrayList <- vector("list", TotalArrayNumber);
m <- 0;
for (Array in ProbesetFiles) {
  m <- m + 1;
  zNormalizedData <- paste(InputDirectory, Array, sep="");
  ArrayList[[m]] <- data.frame(read.delim(zNormalizedData, skip=HeaderLength));
}

#Concatenate respective vectors from each element of the ArrayList
IDVec <- as.character(as.vector(
                             ArrayList[[1]][,xIDPos]));
zLogFoldVec <- as.numeric(as.vector(
                          ArrayList[[1]][,MuFoldPos]));
zRMSIntensityVec <- as.numeric(as.vector(
                               ArrayList[[1]][,xMuAmpPos]));
zProbeNumberVec  <- as.numeric(as.vector(
                               ArrayList[[1]][,xProbeNumberPos]));

for (n in 2:TotalArrayNumber) {
  IDVec <- c(IDVec,
                    as.character(as.vector(
                                 ArrayList[[n]][,xIDPos])));
  zLogFoldVec <- c(zLogFoldVec,
                   as.numeric(as.vector(
                              ArrayList[[n]][,MuFoldPos])));
  zRMSIntensityVec <- c(zRMSIntensityVec,
                        as.numeric(as.vector(
                                   ArrayList[[n]][,xMuAmpPos])));
  zProbeNumberVec  <- c(zProbeNumberVec,
                       as.numeric(as.vector(
                                  ArrayList[[n]][,xProbeNumberPos])));
}

#Filter for data that passes the minimum number of probes required
ProbeFilter <- as.logical(zProbeNumberVec >= MinProbes);

IDVec <- IDVec[ProbeFilter];
zLogFoldVec <- zLogFoldVec[ProbeFilter];
zRMSIntensityVec <- zRMSIntensityVec[ProbeFilter];
zProbeNumberVec <- zProbeNumberVec[ProbeFilter];
      
rm(list=ls(pat="^x"));

###############################################################################
#                               MAIN BODY                                     #
###############################################################################

                 #############################################
                 ####     Apply huberM/MAD Estimators     #### 
                 #############################################

DummyFilter <- rep(TRUE, length(IDVec));

zLogFoldHuberReport <- fhuber.summarize(zLogFoldVec, DummyFilter,
                                IDVec, HuberConf, HuberTol);

zRMSHuberReport <- fhuber.summarize(zRMSIntensityVec, DummyFilter,
                                IDVec, HuberConf, HuberTol);


#Match Huber Reports for Intensity data and LogFold Change Data
#(ie if gene is present in only one, then remove it)
zRMSHuberSetFilter <- !as.logical(
                             match(zRMSHuberReport[,3], c(NA, 0), nomatch=0) |
                             as.numeric(as.vector(zRMSHuberReport[,4])) <
                                        MinArrays);

zLogFoldHuberSetFilter <- !as.logical(
                         match(zLogFoldHuberReport[,3], c(NA, 0), nomatch=0) |
                         as.numeric(as.vector(zLogFoldHuberReport[,4])) <
                                    MinArrays);

zTotHuberSetFilter <- as.logical(zRMSHuberSetFilter * zLogFoldHuberSetFilter);

HuberGenes <- as.character(zLogFoldHuberReport[zTotHuberSetFilter, 1]);
            
Intensity <- as.numeric(as.vector(zRMSHuberReport[zTotHuberSetFilter, 2]));

LogFoldChange <- as.numeric(as.vector(
                                  zLogFoldHuberReport[zTotHuberSetFilter, 2]));

LogFoldStndDev <- as.numeric(as.vector(
                                  zLogFoldHuberReport[zTotHuberSetFilter, 3]));

ArrayNumber <- as.numeric(as.vector(zRMSHuberReport[zTotHuberSetFilter, 4]));

StndErr <- LogFoldStndDev/sqrt(ArrayNumber);

rm(list=ls(pat="^z"));
rm(list=ls(pat="^f"));
rm(list=ls(pat="^g"));
      
                 ##############################################
                 ####    CALCULATE ADJ STANDRD DEVIATION   #### 
                 ##############################################

# LOWESS fit data and organize for extraction of LOWESS values
xBGSigmaLowess <- lowess(Intensity, LogFoldStndDev, f=LowessfParam);

BGStDev <- 0;
xLowessFilter <- order(Intensity);
for (j in 1:length(xLowessFilter)) {
  tempBGStDev <- xBGSigmaLowess$y[as.logical(
                                      match(xLowessFilter, j, nomatch=0))]
  BGStDev <- c(BGStDev, tempBGStDev);
}
                                
BGStDev <- BGStDev[2:(length(BGStDev))]

#Calculate Adjusted Stndard Err
#nu is defined to result in a constant lambda for all genes
nu <- 0;
xAdjustedVar <- 0;
for (k in 1:length(ArrayNumber)) {
  if (ArrayNumber[k] >= Lambda) {
    tempnu <- 0;
    } else {
      tempnu <- Lambda - ArrayNumber[k];
    }
  xSigmaSquared <- (tempnu * BGStDev[k]^2 +
       (ArrayNumber[k]-1) * LogFoldStndDev[k]^2)/
           (tempnu + ArrayNumber[k]-2);
    
  xAdjustedVar <- c(xAdjustedVar, xSigmaSquared);
  nu <- c(nu, tempnu);
}
      
xAdjustedVar <- xAdjustedVar[2:(length(xAdjustedVar))];
nu <- nu[2:length(nu)];
      
AdjustedStndDev <- sqrt(xAdjustedVar);
AdjustedStndErr <- sqrt(xAdjustedVar/ArrayNumber);

rm(list=ls(pat="^temp"));
      
              #### Plot Fits Before and After Adjustment  ####
if (PlotResults) {
  
  #Plot to file
  postscript(file=paste(OutputPrefix, "CyberT", "L",
             as.character(Lambda), ".eps", sep=""),  horizontal=FALSE,
             onefile=TRUE);

  chh <-par()$cxy[2];
  chw <-par()$cxy[2];
  par(mar=c(4,5,3,2), omi=c(0.5,0.5,0.5,0.5));

  par(mfrow=c(2,1));

  y0 <- as.numeric(as.vector(LogFoldStndDev));
  ymax <- max(y0);
  x0 <- as.numeric(as.vector(Intensity));
 
  plot(y0 ~ x0, type="p",  main="StndDev", 
     xlab="A", ylab = "StndDev \nLogFoldProbeSets",
     cex= 1.0, xlim=c(0, 15),
     ylim=c(0, ymax),
     pch=16);
  lines(xBGSigmaLowess$x, xBGSigmaLowess$y, col=2);

  y0 <- AdjustedStndDev;
  x0 <- as.numeric(as.vector(Intensity));

  #Best fit after adjustment
  xCorrectedLowess <- lowess(Intensity, AdjustedStndDev, f=LowessfParam);
  
  plot(y0 ~ x0, type="p",  main="Adj Stnd Dev", 
     xlab="A", ylab = "AdjStndDev \nLogFoldProbeSets",
     cex= 1.0, xlim=c(0, 15),
     ylim=c(0, ymax),
     pch=16);
  lines(xBGSigmaLowess$x, xBGSigmaLowess$y, col=2);
  lines(xCorrectedLowess$x, xCorrectedLowess$y, col=3);
 
  dev.off()
}
rm(list=ls(pat="^x"));
rm(list=ls(pat="^y"));


           ###################################################
           ####    PERFORM t-test AND CALCULATE pVALUES   #### 
           ###################################################

#Calculate t-Test statistic using a pairwise t-Test with log fold data
tStat <- abs(LogFoldChange/AdjustedStndErr);

#Calculate pValue with Bonferroni Correction for comparison
DegreesOfFreedom <- ArrayNumber + nu - 1;
BonferronipValue <- length(tStat) * (2 * (1 - pt(tStat, DegreesOfFreedom)));
BonferronipValue[which(BonferronipValue > 1)] <- 1;

              #####    Perform Permutation Resampling  ####
#All data will be referenced based on relative rank determined by t-test
#statistic
tStatFilter <- order(tStat, decreasing=TRUE);
RankedtStat <- tStat[tStatFilter];

BStat <- rep(0, length(tStat));
B <- TotalPermutations;      
for (b in 1:B) {

  #Create Vector of Resampled LogFoldChange Values
  tempLogFold <- as.numeric(as.vector(ArrayList[[1]][,MuFoldPos]));
  
  xindex <- 1:length(tempLogFold);
  PerFilter <- sample(xindex);
  
  PerLogFold <- tempLogFold[PerFilter];
  
  for (n in 2:TotalArrayNumber) {
    tempLogFold <- as.numeric(as.vector(ArrayList[[n]][,MuFoldPos]));
       
    xindex <- 1:length(tempLogFold);
    PerFilter <- sample(xindex);
    
    PerLogFold <- c(PerLogFold, tempLogFold[PerFilter]);
   
  }

  PerLogFold <- PerLogFold[ProbeFilter];
  
  #Calculate Resampled Median Log Fold for each gene
  PerMedLogFold <- 0;
  for (j in 1:length(HuberGenes)) {
    tempPerMedLogFold <- median(PerLogFold[as.logical(
                                  match(IDVec,
                                        HuberGenes[j], nomatch=0))]);
    PerMedLogFold <- c(PerMedLogFold, tempPerMedLogFold);
  }
  PerMedLogFold <- PerMedLogFold[2:(length(HuberGenes)+1)];

  #Calculate Resample Standard Deviation for each gene
  PerStndDev <- 0;
  for (j in 1:length(HuberGenes)) {
    tempPerStndDev <- sd(PerLogFold[as.logical(
                                  match(IDVec,
                                        HuberGenes[j], nomatch=0))]);
    PerStndDev <- c(PerStndDev, tempPerStndDev);
  }
  PerStndDev <- PerStndDev[2:(length(HuberGenes)+1)];

  #Calculate Adjusted Standard Error for each gene based on distribution of
  #of standard deviation of the actual (ie non-Resampled) data
  PerAdjustedVar <- 0;

  for (k in 1:length(ArrayNumber)) {
    xSigmaSquared <- (nu * BGStDev[k]^2 +
         (ArrayNumber[k]-1) * PerStndDev[k]^2)/
             (nu + ArrayNumber[k]-2);
    PerAdjustedVar <- c(PerAdjustedVar, xSigmaSquared);
  }
  PerAdjustedVar <- PerAdjustedVar[2:(length(ArrayNumber) + 1)];
  PerAdjStndErr <- sqrt(PerAdjustedVar/ArrayNumber);

  #Calculate Bstat (for Adjusted p-values) via Westfall-Young ####
  PertStat <- abs(PerMedLogFold/PerAdjStndErr);
  PertStat <- PertStat[tStatFilter];
  uStat <- RankedtStat[length(RankedtStat)];

  for (j in (length(PertStat)-1):1) {
    NextuStat <- max(uStat[length(PertStat)-j], PertStat[j]);
    uStat <- c(uStat, NextuStat);
  }

  bStat <- rep(0, length(uStat));
  bStat[which(uStat >= RankedtStat)] <- 1;
  BStat <- BStat + bStat;
}  #for (b in 1:B) ie permutation resampling

tempAdjpValue <- BStat/B;

#Stepdown procedure for Adjusted p-values
AdjpValue <- tempAdjpValue[length(tempAdjpValue)];

for (j in (length(tempAdjpValue)-1):1) {
  NextAdjpValue <- max(tempAdjpValue[j], tempAdjpValue[j-1]);
  AdjpValue <- c(AdjpValue, NextAdjpValue);
}

#Reorder AdjpValues from descending to ascending order
#Resulting vector corresponds to ranked t-test statistis
WYpValue <- 1;
for (i in length(AdjpValue):1) {
  WYpValue <- c(WYpValue, AdjpValue[i])
}

WYpValue <- WYpValue[2:length(WYpValue)];

rm(ArrayList)
rm(uStat);
rm(bStat);
rm(list=ls(pat="^x"));
rm(list=ls(pat="^temp"));
rm(list=ls(pat="^Per"));
rm(list=ls(pat="^Next"));

                       ############################
                       #####  OUTPUT RESULTS  #####
                       ############################
      
#Create additional vectors for output
Rank <- c(1:length(tStatFilter));

#Combine vectors into a table for output
CyberTWYTable <- cbind(Rank, ID=HuberGenes[tStatFilter], WYpValue,
                       BonpValue=BonferronipValue[tStatFilter],
                       tStat=tStat[tStatFilter],
                       LogFoldChange=LogFoldChange[tStatFilter],
                       AdjStndErr=AdjustedStndErr[tStatFilter],
                       ArrayNumber=ArrayNumber[tStatFilter],
                       RMSIntensity=Intensity[tStatFilter], BStat, 
                       BGStDev=BGStDev[tStatFilter],
                       StDev=LogFoldStndDev[tStatFilter],
                       StndErr=StndErr[tStatFilter]);

FilterHeader1 <- paste("Input Directory = ", InputDirectory, sep="");
FilterHeader2 <- c("NormFiles = ", paste(as.character(ProbesetFiles),
                                             sep=""));
FilterHeader3 <- date();
FilterHeader4 <- paste("Lambda = ", as.character(Lambda), sep="");
FilterHeader5 <- paste("MinArrays = ", as.character(MinArrays), sep="");
FilterHeader6 <- paste("MinProbes = ", as.character(MinProbes), sep="");   
FilterHeader7 <- paste("LowessfParam = ", as.character(LowessfParam),
                         sep="");          
FilterHeader8 <- c(paste("HuberConf = ", as.character(HuberConf), sep=""),
                   paste("HuberTol = ", as.character(HuberTol), sep=""));
FilterHeader9 <- c("Algorithm = Westfall-Young",
                   paste("Permutation Number = ",
                       as.character(TotalPermutations), sep=""));
FilterHeaderTable <-  c("Rank", "ID", "WYpValue", "BonpValue", "tStat",
                        "LogFoldChng", "AdjStndErr", "ArrayNumber",
                        "Intensity", "BStat", "BGStDev", "StDev", "StndErr");

cat(FilterHeader1, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\n");
cat(FilterHeader2, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader3, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader4, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader5, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader6, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader7, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader8, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeader9, file=paste(OutputPrefix, "L",
                     as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                     append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
                     ".CyTWY", sep=""), sep="\n", append=TRUE);
cat(FilterHeaderTable, file=paste(OutputPrefix, "L",
                         as.character(Lambda), ".CyTWY", sep=""), sep="\t",
                         append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", as.character(Lambda),
             ".CyTWY", sep=""), sep="\n", append=TRUE);

write.table(CyberTWYTable, file=paste(OutputPrefix, "L",
            as.character(Lambda), ".CyTWY", sep=""), quote=FALSE, sep="\t",
            col.names=FALSE, row.names=FALSE, append=TRUE);
            

rm(list=ls());


