#This software is a rerelease of GenomeCrawler.R module of the Genome Crawler
#Suite with alterations in the input and ouput format in order to improve
#transition between modules of eMESS.

#No part of this text or software may be used or reproduced in any matter
#whatsoever without written permission, except in the case of brief quotations
#embodied in critical articles of or for reviews.  For information address
#Brian W. Kirk (bwkbem@yahoo.com).
#Software Copyright (C) June 2003  Brian W. Kirk
#Form/Text Copyright (C) June 2003  Brian W. Kirk
#Released for use with R v1.5.1 (C)

#Copyright (C) June 2003  Brian W. Kirk
#Released for R v1.5.1 (C)

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

###############################################################################
#                                FUNCTIONS                                    #
###############################################################################


###############################################################################
#                           CONSTANT DECLARATION                              #
###############################################################################

#CONSTANTS AND VARIABLES OF TEMPORARY USE BEGIN WITH "x, y, or z" AND ALL
#USER DEFINED FUNCTIONS BEGIN WITH "f" or "g" IN ORDER TO FACILITATE REMOVABLE
#FROM MEMORY.  WHEN USING VARIABLES WITH THESE RESERVED INTIAL LETTERS,
#NEAREST CLEANUP SHOULD BE IDENTIFIED TO INSURE PROPER PERMANENCY


                 ############################################
                 #######  DEFINE PERMANENT CONSTANTS  #######
                 ############################################

#Define Input Directory and File
StatDataIn <- paste(StatDataDirectory, StatDataFile, sep="");
GenomeDataIn <- paste(GenomeDirectory, GenomeFile, sep="");

#Define Output Dirctory with Prefix
OutputPrefix <- paste(OutputDirectory, OutputFile, sep="");


             #####################################################
             ##########  DEFINE TEMPORARY CONSTANTS  #############
             #####################################################

#Data Positions in CyberT output file
xStatRankPos <- 1;
xStatIDPos <- 2;
xStatWYpValuePos  <- 3;
xStatBonpValuePos <- 4;
xStattStatPos <- 5;
xStatLogFoldChngPos <- 6;
xStatStndErrPos <- 7;
xStatArrayNumberPos <- 8;
xStatIntensityPos <- 9;



###############################################################################
#                             VARIABLE DECLARATION                            #
###############################################################################

#CONSTANTS AND VARIABLES OF TEMPORARY USE BEGIN WITH "x", "y", or "z" AND ALL
#USER DEFINED FUNCTIONS BEGIN WITH "f" or "g" IN ORDER TO FACILITATE REMOVABLE
#FROM MEMORY.  WHEN USING VARIABLES WITH THESE RESERVED INTIAL LETTERS,
#NEAREST CLEANUP SHOULD BE IDENTIFIED TO INSURE PROPER PERMANCY


                        #####  Input Tables  ######

#Input StatSigData
xStatDataTable  <- data.frame(read.delim(StatDataIn,
                                       skip=StatHeaderLength));

#Input Ordered Genes from Annotated Genome
xGenomeData <- data.frame(read.delim(GenomeDataIn,
                                       skip=GenomeHeaderLength));


                 #####  Create Vectors from Input Tables  #####

#Data from StatSigTable
zStatRank <- as.numeric(as.vector(xStatDataTable[, xStatRankPos]));
zStatID <- as.character(as.vector(xStatDataTable[,xStatIDPos]));
zStatWYpValue  <- as.numeric(as.vector(xStatDataTable[, xStatWYpValuePos]));
zStatBonpValue <- as.numeric(as.vector(xStatDataTable[, xStatBonpValuePos]));
zStattStat <- as.numeric(as.vector(xStatDataTable[, xStattStatPos]));
zStatLogFoldChng <- as.numeric(as.vector(xStatDataTable[,
                                                        xStatLogFoldChngPos]));
zStatStndErr <- as.numeric(as.vector(xStatDataTable[, xStatStndErrPos]));
zStatArrayNumber <- as.numeric(as.vector(xStatDataTable[,
                                                        xStatArrayNumberPos]));
zStatIntensity <- as.numeric(as.vector(xStatDataTable[, xStatIntensityPos]));

#Find Stat Lambda
xLambdaScan <- scan(file=StatDataIn, what=character(0), sep="\t",
                        nlines=StatHeaderLength);
xLambdaString <- grep("Lambda = ", xLambdaScan,
                          value=TRUE);

if (length(xLambdaString) != 0) {
  StatLambdaFound <- TRUE;
  xStringSplit <- unlist(strsplit(xLambdaString, "= "));
  StatLambda <- as.character(xStringSplit[2]); 
} else {
  StatLambdaFound <- FALSE;
  StatLambda <- "NotFound";
}


if (Bonferroni == TRUE) {
  zStatpValueIn <- zStatBonpValue 
} else {
  zStatpValueIn <- zStatWYpValue
}

#Extract Number of Permutations from Statistical Significance Table Header
if (!Bonferroni) {
  xPermNumberScan <- scan(file=paste(StatDataDirectory, StatDataFile, sep=""),
                       what=character(0), sep="\t", nlines=StatHeaderLength);
  xPermNumberString <- grep("Permutation Number = ", xPermNumberScan,
                          value=TRUE);

  if (length(xPermNumberString) != 0) {
    StatPermNumberFound <- TRUE;
    xStringSplit <- unlist(strsplit(xPermNumberString, "= "));
    StatPermNum <- as.numeric(xStringSplit[2]);
  
  } else {
    StatPermNumberFound <- FALSE;
    StatPermNum <- "NotFound";
  }
} else {
  StatPermNumberFound <- FALSE
}

#Set zero pValues from a permutation algorithm to the minimum value
xpValueSetUpFilter <- as.logical(match(zStatpValueIn, 0, nomatch=0));

if (StatPermNumberFound) {
   zStatpValueIn[xpValueSetUpFilter] <- 1/StatPermNum; 
} else {
  NonZeropValue <- zStatpValueIn[which(zStatpValueIn > 0)];
  minpValue <- min(NonZeropValue);
  
  zStatpValueIn[xpValueSetUpFilter] <- minpValue 
}

       ##### Create Vectors From StatSigData Relative To GenomeGenes #####

GenomeGenes <- as.character(as.vector(xGenomeData[,GenomeIDPos]));
GenomeIndex <- 1:length(GenomeGenes);
GenomeConstant <- length(GenomeGenes);

#Prime Vectors
GenomeRank <- rep("NA", length(GenomeGenes));

GenomepValueIn <- rep(AbsencepValue, length(GenomeGenes));
GenometStat <- rep(AbsenceTestStat, length(GenomeGenes));
GenomeLogFoldChng <- rep(0, length(GenomeGenes));

GenomeStndErr <- rep("NA", length(GenomeGenes));
GenomeIntensity <- rep("NA", length(GenomeGenes));
GenomeArrayNumber <- rep("NA", length(GenomeGenes));

#Substitute Values Corresponding to Genes Present in StatSigData
for (i in 1:length(zStatID)) {
  GeneFilter <- as.logical(match(GenomeGenes, zStatID[i], nomatch=0));
  
  GenomeRank[GeneFilter] <- zStatRank[i];
  GenomepValueIn[GeneFilter] <- zStatpValueIn[i];
  GenometStat[GeneFilter] <- zStattStat[i];

  GenomeLogFoldChng[GeneFilter] <- zStatLogFoldChng[i];
  GenomeStndErr[GeneFilter] <- zStatStndErr[i];
  GenomeIntensity[GeneFilter] <- zStatIntensity[i];
  GenomeArrayNumber[GeneFilter] <- zStatArrayNumber[i];
}

#Index GenomeGenes and Define Query Start and Query End
QueryStartPos <- GenomeIndex[as.logical(match(GenomeGenes, QueryStartID,
                                             nomatch=0))];
QueryEndPos <- GenomeIndex[as.logical(match(GenomeGenes, QueryEndID,
                                             nomatch=0))];

rm(list=ls(pat="^x"));

###############################################################################
#                               MAIN BODY                                     #
###############################################################################


#Search for Statistically Significant Clusters Using a Dynamic Windowing
#Algorithm
ClusterStartName <- "blank";
ClusterEndName <- "blank";
ClusterStartPos <- 0;
ClusterEndPos <- 0;
ClusterpValue <- 1;

for (i in QueryStartPos:(QueryEndPos-WindowLimit)) {
  if (GenomepValueIn[i] < MaxFoldpValue) {
    Search <- 1;
    SearchSize <- WindowLimit;
    while (Search <= SearchSize) { 
      if (GenomepValueIn[(i+Search)] < MaxFoldpValue) {
        tStatSet <- sum(GenometStat[i:(i+Search)]);

        b <- 0;
        NumberOfProbes <- Search + 1;
        for (j in 1:TotalPermutations) {
          PertStatSet <- sum(sample(GenometStat, NumberOfProbes));
          if (PertStatSet > tStatSet) {b <- b+1}  
        }
        AdjpValue <- b/TotalPermutations;
        SearchSize <- SearchSize + 1;

        ClusterStartName <- c(ClusterStartName, GenomeGenes[i]);
        ClusterEndName <- c(ClusterEndName, GenomeGenes[(i+Search)]);
        ClusterpValue <- c(ClusterpValue, AdjpValue);

        tempClusterStartPos <- i;
        tempClusterEndPos <- i+Search;

        ClusterStartPos <- c(ClusterStartPos, tempClusterStartPos);
        ClusterEndPos <- c(ClusterEndPos, tempClusterEndPos);
             
      }
      Search <- Search + 1;
    }
  }
}

ClusterStartName <- ClusterStartName[2:length(ClusterStartName)];
ClusterEndName <- ClusterEndName[2:length(ClusterEndName)];
ClusterStartPos <- ClusterStartPos[2:length(ClusterStartPos)];
ClusterEndPos <- ClusterEndPos[2:length(ClusterEndPos)];

ClusterpValue <- ClusterpValue[2:length(ClusterpValue)];

ClStartGeneName <- ClusterStartName[which(ClusterpValue < MaxClusterpValue)];
ClEndGeneName <- ClusterEndName[which(ClusterpValue < MaxClusterpValue)];
ClStartGenePos <- ClusterStartPos[which(ClusterpValue < MaxClusterpValue)];
ClEndGenePos <- ClusterEndPos[which(ClusterpValue < MaxClusterpValue)];
    
ClClusterProb <- ClusterpValue[which(ClusterpValue < MaxClusterpValue)];


#Use Bayes Theorem to Calculate A Gene Specific Cluster Based pValue and
#Combine this with the LogFold pValue to Calculate an Expression pValue

pValueSetUpFilter <- as.logical(match(ClClusterProb, 0, nomatch=0));
ClClusterProb[pValueSetUpFilter] <- 1/TotalPermutations;

GeGene <- "blank";
GeStartGeneName <- "blank";
GeEndGeneName <- "blank";
GeStartGenePos <- 0;
GeEndGenePos <- 0;

GeExppValue <- 1;

GeFoldpValue <- 1;

GeClusterPosterior <- 1;

GeClusterPrior <- 1;
GeClusterLikelihood <- 1;
GeClusterProb <- 1;

GetStat <- 0
GeLogFoldChng <- 0;
GeStndErr <- 0;
GeIntensity <- 0;
GeArrayNumber <- 0;

for (i in 1:length(ClStartGeneName)) {
  
  GeClusterFront  <-  ClStartGenePos[i];                     
  GeClusterBack <- ClEndGenePos[i];                  
  GeClusterLength <- GeClusterBack - GeClusterFront + 1;

  ClustertStats <- GenometStat[GeClusterFront:GeClusterBack];
  SumClustertStats <- sum(ClustertStats);

  for (j in GeClusterFront:GeClusterBack) {
    BStat <- 0;
    B <- TotalPermutations;
    
    for (b in 1:B) {
      SumPermtStats <- sum(sample(GenometStat, (GeClusterLength-1)));
      
#Prod Mark
      if (SumPermtStats >= (SumClustertStats - GenometStat[j])) {
        BStat <- BStat + 1;
      }

      if (BStat != 0) {
        ClusterLikelihood <- BStat/B
      } else {
        ClusterLikelihood <- 1/B;
      }
    }

    if (GenometStat[j] > 0) {
      ClusterPrior <- as.numeric(as.vector(GenomeRank[j]))/GenomeConstant;
    } else {
      ClusterPrior <- 1;
    }

#    if (ClusterPrior < 1) {
      ClusterPosterior <- ClusterPrior * ClusterLikelihood/ClClusterProb[i]
#    } else {
#      ClusterPosterior <- 1
#    }
    
    ExppValue <- GenomepValueIn[j] * ClusterPosterior;
    
    GeGene <- c(GeGene, GenomeGenes[j]);
    GeStartGeneName <- c(GeStartGeneName, GenomeGenes[GeClusterFront]);
    GeEndGeneName <- c(GeEndGeneName, GenomeGenes[GeClusterBack]);
    GeStartGenePos <- c(GeStartGenePos, ClusterStartPos[i]);
    GeEndGenePos <- c(GeEndGenePos, ClusterEndPos[i]);

    GeExppValue <- c(GeExppValue, ExppValue);
    
    GeFoldpValue <- c(GeFoldpValue, GenomepValueIn[j]);

    GeClusterPosterior <- c(GeClusterPosterior, ClusterPosterior);
    
    GeClusterPrior <- c(GeClusterPrior, ClusterPrior);
    GeClusterLikelihood <- c(GeClusterLikelihood, ClusterLikelihood); 
    GeClusterProb <- c(GeClusterProb, ClClusterProb[i]);

    GetStat <- c(GetStat, GenometStat[j]);
    GeLogFoldChng <- c(GeLogFoldChng,  GenomeLogFoldChng[j]);
    GeStndErr <- c(GeStndErr, GenomeStndErr[j]);
    GeIntensity <- c(GeIntensity, GenomeIntensity[j]);
    GeArrayNumber <- c(GeArrayNumber, GenomeArrayNumber[j]);
  }   
}



GeGene <- GeGene[2:length(GeGene)];
GeStartGeneName <- GeStartGeneName[2:length(GeStartGeneName)];
GeEndGeneName <- GeEndGeneName[2:length(GeEndGeneName)];
GeStartGenePos <- GeStartGenePos[2:length(GeStartGenePos)];
GeEndGenePos <- GeEndGenePos[2:length(GeEndGenePos)];

GeExppValue <- GeExppValue[2:length(GeExppValue)];

GeFoldpValue <- GeFoldpValue[2:length(GeFoldpValue)];

GeClusterPosterior <- GeClusterPosterior[2:length(GeClusterPosterior)];

GeClusterPrior <- GeClusterPrior[2:length(GeClusterPrior)];
GeClusterLikelihood <- GeClusterLikelihood[2:length(GeClusterLikelihood)];
GeClusterProb <-  GeClusterProb[2:length(GeClusterProb)];

GetStat <- GetStat[2:length(GetStat)];
GeLogFoldChng <- GeLogFoldChng[2:length(GeLogFoldChng)];
GeStndErr <- GeStndErr[2:length(GeStndErr)];
GeIntensity <- GeIntensity[2:length(GeIntensity)];
GeArrayNumber <- GeArrayNumber[2:length(GeArrayNumber)];


                       ############################
                       #####  OUTPUT RESULTS  #####
                       ############################
   
#Combine vectors into a table for output
ClusterData <- data.frame(ClStartGeneName, ClEndGeneName, ClClusterProb,
                              ClStartGenePos, ClEndGenePos);

GeneData <- data.frame(cbind(GeGene, GeExppValue, GeFoldpValue,
                             GeClusterPosterior, GetStat, GeLogFoldChng,
                             GeStndErr, GeIntensity, GeArrayNumber,
                             GeClusterPrior, GeClusterLikelihood,
                             GeClusterProb, GeStartGeneName, GeEndGeneName,
                             GeStartGenePos, GeEndGenePos));

# Cluster Data
FilterHeader1 <- paste("StatData File = ", StatDataIn, sep="");
FilterHeader2 <- paste("Genome File = ", GenomeDataIn, sep="");
FilterHeader3 <- date();
FilterHeader4 <- c(paste("Query Start = ", as.character(QueryStartID),
                                           sep=""),
                 paste("Query End = ", as.character(QueryEndID), sep=""));
FilterHeader5 <- c(paste("MaxFoldpValue = ", as.character(MaxFoldpValue),
                         sep=""),
                   paste("MaxClusterpValue = ", as.character(MaxClusterpValue),
                         sep=""));
FilterHeader6 <- paste("Window Limit = ", as.character(WindowLimit),
                                           sep="");   
FilterHeader7 <- c(paste("Total Permutations = ",
                         as.character(TotalPermutations), sep=""));          
FilterHeader8 <- c("Stat Data Input:",
                   paste("Lambda = ", as.character(StatLambda), sep=""),
                   paste("Perm Number = ", as.character(StatPermNum), sep=""));
FilterHeader9 <- c(paste("Bonferroni = ",
                         as.character(Bonferroni), sep=""));          
FilterHeaderTable <-  c("StartGene", "EndGene", "ClusterProb",
                              "StartGenePos", "EndGenePos");

cat(FilterHeader1, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n");
cat(FilterHeader2, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader3, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader4, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader5, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader6, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader7, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader8, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeader9, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);
cat(FilterHeaderTable, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayCl", sep=""), sep="\n", append=TRUE);

write.table(ClusterData, file=paste(OutputPrefix,
                               "L", StatLambda, ".BayCl", sep=""),
                               quote=FALSE, sep="\t", col.names=FALSE,
                               row.names=FALSE, append=TRUE);


# Gene Data
FilterHeader1 <- paste("StatData File = ", StatDataIn, sep="");
FilterHeader2 <- paste("Genome File = ", GenomeDataIn, sep="");
FilterHeader3 <- date();
FilterHeader4 <- c(paste("Query Start = ", as.character(QueryStartID),
                                           sep=""),
                 paste("Query End = ", as.character(QueryEndID), sep=""));
FilterHeader5 <- c(paste("MaxFoldpValue = ", as.character(MaxFoldpValue),
                         sep=""),
                   paste("MaxClusterpValue = ", as.character(MaxClusterpValue),
                         sep=""));
FilterHeader6 <- paste("Window Limit = ", as.character(WindowLimit),
                                           sep="");   
FilterHeader7 <- c(paste("Total Permutations = ",
                         as.character(TotalPermutations), sep=""));          
FilterHeader8 <- c("Stat Param:",
                   paste("Lambda = ", as.character(StatLambda), sep=""),
                   paste("Perm Number = ", as.character(StatPermNum), sep=""));
FilterHeader9 <- c(paste("Bonferroni = ",
                         as.character(Bonferroni), sep=""));          
FilterHeaderTable <-  c("Gene", "Exp pValue", "Fold pValue", "Posterior",
                        "tStat", "LogFoldChng", "StndErr", "Intensity",
                        "ArrayNumber", "Prior", "Likelihood", "Clust Prob",
                        "ClusterStart", "ClusterEnd", "ClStartPos",
                        "ClEndPos");

cat(FilterHeader1, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n");
cat(FilterHeader2, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader3, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader4, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader5, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader6, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader7, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader8, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeader9, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);
cat(FilterHeaderTable, file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\t", append=TRUE);
cat(c(""), file=paste(OutputPrefix, "L", StatLambda,
                     ".BayGe", sep=""), sep="\n", append=TRUE);

write.table(GeneData, file=paste(OutputPrefix, 
                               "L", StatLambda, ".BayGe", sep=""),
                               quote=FALSE, sep="\t", col.names=FALSE,
                               row.names=FALSE, append=TRUE);

rm(list=ls());
